typedef long int64t;

inline void accumBilateral( const int64_t delta, const int64_t i, const int64_t j, int64_t* A, int64_t* b){
    int64_t f = (abs(delta) > 30 )? 0:1;
    A[0] += f*i*i; A[1] += f*i*j; A[3] += f*j*j;
    b[0] += f*i*delta;
    b[1] += f*j*delta;
}

template <int r>
void bilateralFilter( int W, int H, const uint16_t* depth, float* normals )
{
    const int offset0 = -r-r*W; const int offset1 = 0-r*W; const int offset2 = +r-r*W;
    const int offset3 = -r;                                const int offset4 = +r;
    const int offset5 = -r+r*W; const int offset6 = 0+r*W; const int offset7 = +r+r*W;


    for(int y=r;y<H-r-1;++y) {
        const uint16_t* line = depth  + y*W + r;
        float*             n = normals+ 3*(y*W + r);
        for(int x=r;x<W-r-1;++x) {
            int64_t d = line[0];
            // accum
            int64_t A[4]; A[0] = A[1] = A[2] = A[3] = 0;
            int64_t b[2]; b[0] = b[1] = 0;
            accumBilateral( line[offset0] - d, -r, -r, A, b);
            accumBilateral( line[offset1] - d,  0, -r, A, b);
            accumBilateral( line[offset2] - d, +r, -r, A, b);
            accumBilateral( line[offset3] - d, -r,  0, A, b);
            accumBilateral( line[offset4] - d, +r,  0, A, b);
            accumBilateral( line[offset5] - d, -r, +r, A, b);
            accumBilateral( line[offset6] - d,  0, +r, A, b);
            accumBilateral( line[offset7] - d, +r, +r, A, b);
            // solve
            int64_t det =  A[0]*A[3] - A[1]*A[1];
            int64_t ddx =  A[3]*b[0] - A[1]*b[1];
            int64_t ddy = -A[1]*b[0] + A[0]*b[1];
            float nx = 1150*ddx;
            float ny = 1150*ddy;
            float nz = -det*d;
            float norminv = 1.0/sqrtf(nx*nx+ny*ny+nz*nz);
            n[0] = nx*norminv;
            n[1] = ny*norminv;
            n[2] = nz*norminv;
            line++;
            n+=3;
        }
    }
}